@@ -5,104 +5,102 @@
 #  /_/\_\|_|  \_/\_/  |_||_| |_| \__, ||___/ _____  \__,_| \__||_||_||___/
 #                                |___/      |_____|
 
-__version__ = "25.0.6"
+__version__ = "25.1.2"
 
 
-import dropbox
 from pathlib import Path
 import os
 import sys
 import math
 import base64
+import requests
+import json
 
-dbx = None
 Pythonista = sys.platform == "ios"
+
 try:
     import xlwings
 
     xlwings = True
+    import pyodide_http
+
 except ImportError:
     xlwings = False
 
 missing = object()
 
+_token = None
+
 
 def dropbox_init(refresh_token=missing, app_key=missing, app_secret=missing, **kwargs):
     """
-    dropbox initialize
-
     This function may to be called prior to using any dropbox function
     to specify the request token, app key and app secret.
-    If these are specified as REFRESH_TOKEN, APP_KEY and APP_SECRET
-    environment variables, it is no necessary to call dropbox_init().
+    If these are specified as DROPBOX.REFRESH_TOKEN, DROPBOX.APP_KEY and DROPBOX.APP_SECRET
+    environment variables, it is not necessary to call dropbox_init().
 
     Parameters
     ----------
     refresh_token : str
         oauth2 refreshntoken
 
-        if omitted: use the environment variable REFRESH_TOKEN
+        if omitted: use the environment variable DROPBOX.REFRESH_TOKEN
 
     app_key : str
         app key
 
-        if omitted: use the environment variable APP_KEY
+        if omitted: use the environment variable DROPBOX.APP_KEY
 
 
     app_secret : str
         app secret
 
-        if omitted: use the environment variable APP_SECRET
+        if omitted: use the environment variable DROPBOX.APP_SECRET
 
     Returns
     -------
-    -
+    dropbox object
     """
-    global dbx
 
-    if Pythonista:
-        # under Pythonista, the environ is updated from the .environ.toml file, if present
-        environ_file = Path(os.environ["HOME"]) / "Documents" / ".environ.toml"
-
-        if environ_file.is_file():
-            with open(environ_file, "r") as f:
-                import toml
-
-                d = toml.load(f)
-                os.environ.update(d)
+    global _token
+    if xlwings:
+        pyodide_http.patch_all()  # required to reliably use requests on pyodide platforms
 
     if refresh_token is missing:
-        if "REFRESH_TOKEN" in os.environ:
-            refresh_token = os.environ["REFRESH_TOKEN"]
+        if "DROPBOX.REFRESH_TOKEN" in os.environ:
+            refresh_token = os.environ["DROPBOX.REFRESH_TOKEN"]
         else:
-            raise ValueError("no REFRESH_TOKEN found in environment.")
+            raise ValueError("no DROPBOX.REFRESH_TOKEN found in environment.")
     if app_key is missing:
-        if "APP_KEY" in os.environ:
-            app_key = os.environ["APP_KEY"]
+        if "DROPBOX.APP_KEY" in os.environ:
+            app_key = os.environ["DROPBOX.APP_KEY"]
         else:
-            raise ValueError("no APP_KEY found in environment.")
+            raise ValueError("no DROPBOX.APP_KEY found in environment.")
     if app_secret is missing:
-        if "APP_SECRET" in os.environ:
-            app_secret = os.environ["APP_SECRET"]
+        if "DROPBOX.APP_SECRET" in os.environ:
+            app_secret = os.environ["DROPBOX.APP_SECRET"]
         else:
-            raise ValueError("no APP_SECRET found in environment.")
+            raise ValueError("no DROPBOX.APP_SECRET found in environment.")
 
-    dbx = dropbox.Dropbox(oauth2_refresh_token=refresh_token, app_key=app_key, app_secret=app_secret, **kwargs)
+    response = requests.post(
+        "https://api.dropbox.com/oauth2/token",
+        data={"grant_type": "refresh_token", "refresh_token": refresh_token, "client_id": app_key, "client_secret": app_secret},
+        timeout=30,
+    )
     try:
-        dbx.files_list_folder(path="")  # just to test proper credentials
-    except dropbox.exceptions.AuthError:
+        response.raise_for_status()
+    except requests.exceptions.HTTPError:
         raise ValueError("invalid dropbox credentials")
+    _token = response.json()["access_token"]
 
 
-def _login_dbx():
-    if dbx is None:
+def _login_dropbox():
+    if _token is None:
         dropbox_init()  # use environment
 
 
 def list_dropbox(path="", recursive=False, show_files=True, show_folders=False):
     """
-    list_dropbox
-
     returns all dropbox files/folders in path
 
     Parameters
@@ -111,7 +109,7 @@ def list_dropbox(path="", recursive=False, show_files=True, show_folders=False):
         path from which to list all files (default: '')
 
     recursive : bool
-        if True, recursively list files. if False (default) no recursion
+        if True, recursively list files and folders. if False (default) no recursion
 
     show_files : bool
         if True (default), show file entries
@@ -125,50 +123,44 @@ def list_dropbox(path="", recursive=False, show_files=True, show_folders=False):
     -------
     files : list
 
-    Note
-    ----
-    Directory entries are never returned
-
     Note
     ----
     If REFRESH_TOKEN, APP_KEY and APP_SECRET environment variables are specified,
     it is not necessary to call dropbox_init() prior to any dropbox function.
     """
-    _login_dbx()
-    out = []
-    result = dbx.files_list_folder(path, recursive=recursive)
+    _login_dropbox()
 
-    for entry in result.entries:
-        if show_files and isinstance(entry, dropbox.files.FileMetadata):
-            out.append(entry.path_display)
-        if show_folders and isinstance(entry, dropbox.files.FolderMetadata):
-            out.append(entry.path_display + "/")
+    API_RPC = "https://api.dropboxapi.com/2"
+    headers = {"Authorization": f"Bearer {_token}", "Content-Type": "application/json"}
+    payload = {"path": str(path), "recursive": recursive, "include_deleted": False}
+    response = requests.post("https://api.dropboxapi.com/2/files/list_folder", headers=headers, json=payload, timeout=30)
+    response.raise_for_status()
+    data = response.json()
+    entries = data["entries"]
+    while data.get("has_more"):
+        response = requests.post(f"{API_RPC}/files/list_folder/continue", headers=headers, json={"cursor": data["cursor"]}, timeout=30)
+        response.raise_for_status()
+        data = response.json()
+        entries.extend(data["entries"])
 
-    while result.has_more:
-        result = dbx.files_list_folder_continue(result.cursor)
-        for entry in result.entries:
-            if show_files and isinstance(entry, dropbox.files.FileMetadata):
-                out.append(entry.path_display)
-            if show_folders and isinstance(entry, dropbox.files.FolderMetadata):
-                out.append(entry.path_display + "/")
-
-    return out
+    result = []
+    for entry in entries:
+        if show_files and entry[".tag"] == "file":
+            result.append(entry["path_display"])
+        if show_folders and entry[".tag"] == "folder":
+            result.append(entry["path_display"] + "/")
+    return result
 
 
-def read_dropbox(dropbox_path, max_retries=100):
+def read_dropbox(dropbox_path):
     """
-    read_dropbox
-
-    read from dopbox at given path
+    read file from dropbox
 
     Parameters
     ----------
     dropbox_path : str or Pathlib.Path
         path to read from
 
-    max_retries : int
-        number of retries (default: 100)
-
     Returns
     -------
     contents of the dropbox file : bytes
@@ -177,25 +169,25 @@ def read_dropbox(dropbox_path, max_retries=100):
     ----
     If REFRESH_TOKEN, APP_KEY and APP_SECRET environment variables are specified,
     it is not necessary to call dropbox_init() prior to any dropbox function.
-
-    As reading from dropbox is very unreliable under pyodide, reading will have to be retried (by default maximum 100 times).
-    The number of retries can be found with read_dropbox.retries.
     """
 
-    _login_dbx()
-    for read_dropbox.retries in range(max_retries + 1):
-        metadata, response = dbx.files_download(dropbox_path)
-        file_content = response.content
-        if len(file_content) == metadata.size:
-            return file_content
-    raise OSError(f"after {max_retries} still no valid response")
+    _login_dropbox()
+    headers = {"Authorization": f"Bearer {_token}", "Dropbox-API-Arg": json.dumps({"path": dropbox_path})}
+    with requests.post("https://content.dropboxapi.com/2/files/download", headers=headers, stream=True, timeout=60) as response:
+        try:
+            response.raise_for_status()
+        except requests.exceptions.HTTPError as e:
+            raise FileNotFoundError(f"file {str(dropbox_path)} not found. Original message is {e}") from None
+        chunks = []
+        for chunk in response.iter_content(chunk_size=1024):
+            if chunk:
+                chunks.append(chunk)
+    return b"".join(chunks)
 
 
 def write_dropbox(dropbox_path, contents):
     """
-    write_dropbox
-
-    write from dopbox at given path
+    write to file on dropbox
 
     Parameters
     ----------
@@ -210,15 +202,47 @@ def write_dropbox(dropbox_path, contents):
     If REFRESH_TOKEN, APP_KEY and APP_SECRET environment variables are specified,
     it is not necessary to call dropbox_init() prior to any dropbox function.
     """
-    _login_dbx()
-    dbx.files_upload(contents, dropbox_path, mode=dropbox.files.WriteMode.overwrite)
+    _login_dropbox()
+    headers = {
+        "Authorization": f"Bearer {_token}",
+        "Dropbox-API-Arg": json.dumps(
+            {"path": str(dropbox_path), "mode": "overwrite", "autorename": False, "mute": False}  # Where it will be saved in Dropbox  # "add" or "overwrite"
+        ),
+        "Content-Type": "application/octet-stream",
+    }
+    response = requests.post("https://content.dropboxapi.com/2/files/upload", headers=headers, data=contents)
+    return response
+
+
+def delete_from_dropbox(dropbox_path):
+    """
+    delete file dropbox
+
+    Parameters
+    ----------
+    dropbox_path : str or Pathlib.Path
+        path to delete
+
+    Note
+    ----
+    If REFRESH_TOKEN, APP_KEY and APP_SECRET environment variables are specified,
+    it is not necessary to call dropbox_init() prior to any dropbox function.
+    """
+    _login_dropbox()
+
+    headers = {"Authorization": f"Bearer {_token}", "Content-Type": "application/json"}
+
+    data = {"path": str(dropbox_path)}  # Path in Dropbox, starting with /
+
+    response = requests.post("https://api.dropboxapi.com/2/files/delete_v2", headers=headers, data=json.dumps(data))
+    if response.status_code == 200:
+        return response
+    raise FileNotFoundError(f"dropbox file {dropbox_path} not found")
 
 
 def list_local(path, recursive=False, show_files=True, show_folders=False):
     """
-    list_local
-
-    returns all local files/folders in path
+    returns all local files/folders at given path
 
     Parameters
     ----------
@@ -276,91 +300,215 @@ class block:
     Parameters
     ----------
     number_of_rows : int
-        number of rows
+        number of rows (dedault 1)
 
     number_of_columns : int
-        number of columns
+        number of columns (default 1)
 
     Returns
     -------
     block
     """
 
-    def __init__(self, value=missing, *, number_of_rows=missing, number_of_columns=missing, column_like=False):
+    def __init__(self, number_of_rows=1, number_of_columns=1):
         self.dict = {}
-        self.column_like = column_like
-        if value is missing:
-            if number_of_rows is missing:
-                number_of_rows = 1
-            if number_of_columns is missing:
-                number_of_columns = 1
-            self.number_of_rows = number_of_rows
-            self.number_of_columns = number_of_columns
+        self.number_of_rows = number_of_rows
+        self.number_of_columns = number_of_columns
+        self._invalidate_highest_used_cache()
 
-        else:
-            if isinstance(value, block):
-                value = value.value
-            self.value = value
-            if number_of_rows is not missing:
-                self.number_of_rows = number_of_rows
-            if number_of_columns is not missing:
-                self.number_of_columns = number_of_columns
+    def __eq__(self, other):
+        if isinstance(other, block):
+            return self.value == other.value
+        return False
 
     @classmethod
-    def from_xlrd_sheet(cls, sheet, number_of_rows=missing, number_of_columns=missing):
+    def from_value(cls, value, column_like=False):
+        """
+        makes a block from a given value
+
+        Parameters
+        ----------
+        value : scalar, list of scalars, list of lists of scalars or block
+            value to be used in block, possibly expanded to a list of lists of scalars
+
+        column_like : boolean
+            if value is a list of scalars, values is interpreted as a column if True, as a row otherwise
+
+        Returns
+        -------
+        block : block
+        """
+        if isinstance(value, block):
+            value = value.value
+        if not isinstance(value, list):
+            value = [[value]]
+        if not isinstance(value[0], list):
+            if column_like:
+                value = [[item] for item in value]
+            else:
+                value = [value]
+        bl = cls(len(value), 1)
+
+        for row, row_contents in enumerate(value, 1):
+            for column, item in enumerate(row_contents, 1):
+                if item and not (isinstance(item, float) and math.isnan(item)):
+                    bl.dict[row, column] = item
+                bl._number_of_columns = max(bl.number_of_columns, column)
+        return bl
+
+    @classmethod
+    def from_range(cls, rng):
+        """
+        makes a block from a given range
+
+        Parameters
+        ----------
+        rng : xlwings.Range
+            range to be used be used in block
+
+        Returns
+        -------
+        block : block
+        """
+        number_of_rows, number_of_columns = rng.shape
+        return cls.from_value(rng.value, column_like=(number_of_columns == 1))
+
+    @classmethod
+    def from_xlrd_sheet(cls, sheet):
+        """
+        makes a block from a xlrd sheet
+
+        Parameters
+        ----------
+        sheet : xlrd sheet
+            sheet to be used be used in block
+
+        Returns
+        -------
+        block : block
+        """
         v = [sheet.row_values(row_idx)[0 : sheet.ncols] for row_idx in range(0, sheet.nrows)]
-        return cls(v, number_of_rows=number_of_rows, number_of_columns=number_of_columns)
+        return cls.from_value(v)
 
     @classmethod
-    def from_openpyxl_sheet(cls, sheet, number_of_rows=missing, number_of_columns=missing):
+    def from_openpyxl_sheet(cls, sheet):
+        """
+        makes a block from an openpyxl sheet
+
+        Parameters
+        ----------
+        sheet : xlrd sheet
+            sheet to be used be used in block
+
+        Returns
+        -------
+        block : block
+        """
         v = [[cell.value for cell in row] for row in sheet.iter_rows()]
-        return cls(v, number_of_rows=number_of_rows, number_of_columns=number_of_columns)
+        return cls.from_value(v)
 
     @classmethod
-    def from_file(cls, filename, number_of_rows=missing, number_of_columns=missing):
+    def from_file(cls, filename):
+        """
+        makes a block from a file
+
+        Parameters
+        ----------
+        filename : str
+            file to be used be used in block
+
+        Returns
+        -------
+        block : block
+        """
         with open(filename, "r") as f:
             v = [[line if line else missing] for line in f.read().splitlines()]
-        return cls(v, number_of_rows=number_of_rows, number_of_columns=number_of_columns)
+        return cls.from_value(v)
 
     @classmethod
-    def from_dataframe(cls, df, number_of_rows=missing, number_of_columns=missing):
+    def from_dataframe(cls, df):
+        """
+        makes a block from a given dataframe
+
+        Parameters
+        ----------
+        df : pandas dataframe
+            dataframe to be used be used in block
+
+        Returns
+        -------
+        block : block
+        """
         v = df.values.tolist()
-        return cls(v, number_of_rows=number_of_rows, number_of_columns=number_of_columns)
+        return cls.from_value(v)
 
     def to_openpyxl_sheet(self, sheet):
+        """
+        appends a block to a given openpyxl sheet
+
+        Parameters
+        ----------
+        sheet: openpyxl sheet
+            sheet to be used be used
+
+        Returns
+        -------
+        block : block
+        """
         for row in self.value:
             sheet.append(row)
 
+    def reshape(self, number_of_rows=missing, number_of_columns=missing):
+        """
+        makes a new block with given dimensions
+
+        Parameters
+        ----------
+        number_of_rows : int
+            if given, expand or shrink to the given number of rows
+
+        number_of_columns : int
+            if given, expand or shrink to the given number of columns
+
+        Returns
+        -------
+        block : block
+        """
+        if number_of_rows is missing:
+            number_of_rows = self.number_of_rows
+        if number_of_columns is missing:
+            number_of_columns = self.number_of_columns
+        bl = block(number_of_rows=number_of_rows, number_of_columns=number_of_columns)
+        for (row, column), value in self.dict.items():
+            if row <= number_of_rows and column <= number_of_columns:
+                bl[row, column] = value
+        return bl
+
     @property
     def value(self):
         return [[self.dict.get((row, column)) for column in range(1, self.number_of_columns + 1)] for row in range(1, self.number_of_rows + 1)]
 
-    @value.setter
-    def value(self, value):
-        if not isinstance(value, list):
-            value = [[value]]
-        if not isinstance(value[0], list):
-            if self.column_like:
-                value = [[item] for item in value]
-            else:
-                value = [value]
-
-        self.number_of_rows = len(value)
-        self._number_of_columns = 0
-
-        for row, row_contents in enumerate(value, 1):
-            for column, item in enumerate(row_contents, 1):
-                if item and not (isinstance(item, float) and math.isnan(item)):
-                    self.dict[row, column] = item
-                    self._number_of_columns = max(self.number_of_columns, column)
+    def _invalidate_highest_used_cache(self):
+        self._highest_used_row_number = None
+        self._highest_used_column_number = None
 
     def __setitem__(self, row_column, value):
         row, column = row_column
         if row < 1 or row > self.number_of_rows:
-            raise IndexError(f"row must be between 1 and {self.number_of_rows} not {row}")
+            raise IndexError(f"row must be between 1 and {self.number_of_rows}; not {row}")
         if column < 1 or column > self.number_of_columns:
-            raise IndexError(f"column must be between 1 and {self.number_of_columns} not {column}")
-        self.dict[row, column] = value
+            raise IndexError(f"column must be between 1 and {self.number_of_columns}; not {column}")
+        if value is None:
+            if (row, column) in self.dict:
+                del self.dict[row, column]
+                self._invalidate_highest_used_cache()
+
+        else:
+            self.dict[row, column] = value
+            if self._highest_used_row_number:
+                self._highest_used_row_number = max(self._highest_used_row_number, row)
+            if self._highest_used_column_number:
+                self._highest_used_column_number = max(self._highest_used_column_number, column)
 
     def __getitem__(self, row_column):
         row, column = row_column
@@ -377,7 +525,7 @@ class block:
         minimized block : block
              uses highest_used_row_number and highest_used_column_number to minimize the block
         """
-        return block(self, number_of_rows=self.highest_used_row_number, number_of_columns=self.highest_used_column_number)
+        return self.reshape(number_of_rows=self.highest_used_row_number, number_of_columns=self.highest_used_column_number)
 
     @property
     def number_of_rows(self):
@@ -386,7 +534,8 @@ class block:
     @number_of_rows.setter
     def number_of_rows(self, value):
         if value < 1:
-            raise ValueError(f"number_of_rows should be >=1, not {value}")
+            raise ValueError(f"number_of_rows should be >=1; not {value}")
+        self._invalidate_highest_used_cache()
         self._number_of_rows = value
         for row, column in list(self.dict):
             if row > self._number_of_rows:
@@ -399,7 +548,8 @@ class block:
     @number_of_columns.setter
     def number_of_columns(self, value):
         if value < 1:
-            raise ValueError(f"number_of_columns should be >=1, not {value}")
+            raise ValueError(f"number_of_columns should be >=1; not {value}")
+        self._invalidate_highest_used_cache()
         self._number_of_columns = value
         for row, column in list(self.dict):
             if column > self._number_of_columns:
@@ -407,17 +557,22 @@ class block:
 
     @property
     def highest_used_row_number(self):
-        if self.dict:
-            return max(row for (row, column) in self.dict)
-        else:
-            return 1
+        if not self._highest_used_row_number:
+            if self.dict:
+                self._highest_used_row_number = max(row for (row, column) in self.dict)
+            else:
+                self._highest_used_row_number = 1
+        return self._highest_used_row_number
 
     @property
     def highest_used_column_number(self):
-        if self.dict:
-            return max(column for (row, column) in self.dict)
-        else:
-            return 1
+        if not self._highest_used_column_number:
+            if self.dict:
+                self._highest_used_column_number = max(column for (row, column) in self.dict)
+            else:
+                self._highest_used_column_number = 1
+
+        return self._highest_used_column_number
 
     def __repr__(self):
         return f"block({self.value})"
@@ -434,7 +589,7 @@ class block:
         if column > self.number_of_columns:
             raise ValueError(f"{name}={column} > number_of_columns={self.number_of_columns}")
 
-    def transpose(self):
+    def transposed(self):
         """
         transpose block
 
@@ -699,12 +854,21 @@ class block:
 
     def decode_to_files(self):
         """
-        decode the block with encoded file(s) to individual pyoidide files
+        decode the block with encoded file(s) to individual pyoidide file(s)
+
+        Returns
+        -------
+        count : int
+            number of files decoded
+
+        Note
+        ----
+        if the block does not contain an encode file, the method just returns 0
         """
         count = 0
         for column in range(1, self.number_of_columns + 1):
             row = 1
-            bl = self.mimimized()
+            bl = self.minimized()
             while row <= self.number_of_rows:
                 if self[row, column] and self[row, column].startswith("<file=") and self[row, column].endswith(">"):
                     filename = self[row, column][6:-1]
@@ -875,3 +1039,4 @@ def trigger_macro(sheet):
 
 if __name__ == "__main__":
     ...
+
